local Collider = require "collider"
local shapes = require "collider.shapes"

local gui = require "quickie"

local Class = require "hump.class"
local Vector = require "hump.vector"

local hsv = require "hsv"

local Operator = Class {}

function Operator:init(name, code, apply)
	self.name = name
	self.code = code
	self.apply = apply
end

function Operator:compile(inputs)
	local operands = {}

	for input in pairs(inputs) do
		table.insert(operands, input:compile())
	end

	for i=#operands+1,2 do
		operands[i] = "0"
	end

	local arguments = operands[1] .. ", " .. operands[2]

	return self.name .. "(" .. arguments .. ")"
end

local Expression = Class {}

function Expression:init(operator, operands)
	self.operator = operator
	self.operands = operands or {}
end

function Expression:compile()
	local operands = {}

	for i, operand in ipairs(self.operands) do
		operands[i] = operand:compile()
	end

	local arguments = table.concat(operands, ", ")

	return self.operator.name .. "(" .. arguments .. ")"
end

function Expression:compute(...)
	local operands = {}

	for i, operand in ipairs(self.operands) do
		operands[i] = operand:compute(...)
	end

	return self.operator:apply(unpack(operands))
end

local Variable = Class {}

function Variable:init(variable)
	self.name = variable
	self.variable = variable
end

function Variable:compute(variables)
	return variables[self.variable]
end

function Variable:compile()
	return self.variable
end

local function mix(a, b, u)
	return a * (1 - u) + b * u
end

local function clamp(min, val, max)
	return math.min(max, math.max(min, val))
end

local VARIABLES = {
	Variable("t"),
	Variable("x"),
	Variable("y"),
--	Variable("noise"),
}

local BINARY = {
	Operator("g_add",
	[[float g_add(float a, float b)
{
	return mod(a + b, 1);
}
	]],
	function (self, a, b) 
		return (a + b) % 1
	end),

	Operator("g_quantise",
	[[float g_quantise(float a, float b)
{
	float q = mix(2, 32, b);
	return floor(a * q) / q;
}
	]],
	function (self, a, b) 
		local q = mix(2, 8, b)
		return math.floor(a * q) / q
	end),
	Operator("g_sin",
	[[float g_sin(float a, float b)
{
	float f = pow(2, -b * 4);
	return sin(a * M_PI * 2 * f) * 0.5 + 0.5;
}
	]],
	function (self, a, b) 
		local f = math.pow(2, -b * 4)
		return math.sin(a * math.pi * 2 * f) * 0.5 + 0.5
	end),
	Operator("g_mul",
	[[float g_mul(float a, float b)
{
	return a * b;
}
	]],
	function (self, a, b) 
		return a * b
	end),
	Operator("g_mod",
	[[float g_mod(float a, float b)
{
	return mod(a, b);
}
	]],
	function (self, a, b) 
		return a % b
	end),
	Operator("g_acos",
	[[float g_acos(float a, float b)
{
	return b * acos(a * 2 - 1) / (M_PI * 2);
}
	]],
	function (self, a, b) 
		return b * math.acos(a * 2 - 1) / (math.pi * 2)
	end),
}

function random_expression(variables, operators)
	local nodes = {}

	table.insert(nodes, variables[1])

	for i=1,math.random(16,32) do
		table.insert(nodes, variables[math.random(#variables)])
	end

	for i=1,math.random(16,32) do
		table.insert(nodes, Variable(tostring(math.random())))
	end

	while #nodes > 1 do
		local operator = operators[math.random(#operators)]
		local operand1 = table.remove(nodes, math.random(#nodes))
		local operand2 = table.remove(nodes, math.random(#nodes))

		table.insert(nodes, Expression(operator, {operand1, operand2}))
	end

	return nodes[1]
end

local Glitch = Class {}

function Glitch:init()
	self.expressions = {
		random_expression(VARIABLES, BINARY),
		random_expression(VARIABLES, BINARY),
		random_expression(VARIABLES, BINARY),

		random_expression(VARIABLES, BINARY),
		random_expression(VARIABLES, BINARY),
	}
end

function Glitch:compile(r, g, b)
	local hsv = love.filesystem.read("hsv.glsl")

	lines = {}

	table.insert(lines, "#define M_PI 3.14159265358979323846")
	table.insert(lines, hsv)

	for i, operator in ipairs(BINARY) do
		table.insert(lines, operator.code)
	end

	table.insert(lines, "extern number t, s, h;")
	table.insert(lines, "vec4 effect(vec4 colour, Image image, vec2 local, vec2 screen)")
	table.insert(lines, "{")
	table.insert(lines, "	float x = screen.x / s; float y = (h - screen.y) / s;")
	table.insert(lines, "	float r = " .. r:compile() .. ";")
	table.insert(lines, "	float g = " .. g:compile() .. ";")
	table.insert(lines, "	float b = " .. b:compile() .. ";")
	table.insert(lines, "	return vec4(vec3(r, g, b), 1.0 - 0.00000000000000000000001*t*x*y);")
	table.insert(lines, "}")

	return table.concat(lines, "\n")
end

function Glitch:compute(variables)
	local r = self.expressions[1]:compute(variables) * 255
	local g = self.expressions[2]:compute(variables) * 255
	local b = self.expressions[3]:compute(variables) * 255

	local r1, g1, b1 = hsv(self.expressions[4]:compute(variables) * 255, 255, 255)
	local r2, g2, b2 = hsv(self.expressions[5]:compute(variables) * 255, 255, 255)

	local r = r * 0.5 + r1 * 0.25 + r2 * 0.25
	local g = g * 0.5 + g1 * 0.25 + g2 * 0.25
	local b = b * 0.5 + b1 * 0.25 + b2 * 0.25

	return r, g, b, 255
end

function Glitch:mutate()
	for i, expression in ipairs(self.expressions) do
		self.expressions[1].operator = BINARY[math.random(#BINARY)]
	end
end

function Rectangle(w, h, xo, yo)
    local x, y = w / 2, h / 2
    local xo, yo = xo or 0, yo or 0

    local points = {
        -x+xo, -y+yo;
         x+xo, -y+yo;
         x+xo,  y+yo;
        -x+xo,  y+yo;
    }

    local shape = shapes.newPolygonShape(unpack(points))

    return shape
end

local Node = Class {
	SIZE = 25,
}

function Node:init(operator, x, y, collider)
	self.operator = operator

	self.shape = shapes.newCircleShape(x, y, self.SIZE)
	self.shape.type = "node"
	self.shape.self = self
	collider:addShape(self.shape)

	self.inputs = {}
	self.outputs = {}
	self.ancestors = {[self] = self}

	print(self.operator.name)
end

function Node:draw()
	local x, y = self:position()

	self.shape:draw("fill")

	love.graphics.setBlendMode("subtractive")
	love.graphics.printf(self.operator.name, x - self.SIZE, y - 6, self.SIZE*2, 'center')
	love.graphics.setBlendMode("alpha")

	if self.outputs then
		for output in pairs(self.outputs) do
			local start = Vector(self:position())
			local stop = Vector(output:position())

			start = start + (stop - start):normalized() * self.SIZE

			love.graphics.circle("fill", start.x, start.y, self.SIZE * 0.2)
			love.graphics.line(start.x, start.y, stop.x, stop.y)
		end
	end
end

function Node:position()
	return self.shape:center()
end

function Node:move(x, y)
	self.shape:moveTo(x, y)
end

function Node:into(node)
	if node == self or not node.inputs then return end
	
	if self.outputs[node] then
		node:unlink(self)
		self.outputs[node] = nil
	else
		node:from(self) 
		self.outputs[node] = node
	end
end

function Node:from(node)
	self.inputs[node] = node

	self:update_ancestors()
end

function Node:unlink(node)
	self.inputs[node] = nil

	self:update_ancestors()
end

function Node:update_ancestors()
	self.ancestors = {[self] = self}

	for input in pairs(self.inputs) do
		for ancestor in pairs(input.ancestors) do
			self.ancestors[ancestor] = ancestor
		end
	end

	if self.outputs then
		for output in pairs(self.outputs) do
			output:update_ancestors()
		end
	end
end

function Node:compile()
	return self.operator:compile(self.inputs)
end

local Terminal = Class { __includes=Node}

function Terminal:init(...)
	Node.init(self, ...)

	self.static = true
	self.outputs = nil
end

function Terminal:from(node)
	for input in pairs(self.inputs) do
		input.outputs[self] = nil
	end

	self.inputs = {}
	self.inputs[node] = node

	self:update_ancestors()
end

function Terminal:compile()
	local formula = "0"

	for input in pairs(self.inputs) do
		formula = input:compile()
	end

	return formula
end

local Input = Class { __includes=Node}

function Input:init(...)
	Node.init(self, ...)

	self.static = true
	self.inputs = nil
end

function Input:compile()
	return self.operator:compile()
end

function refresh()
	size   = math.min(love.graphics.getWidth(), love.graphics.getHeight())
	shader = love.graphics.newPixelEffect(automaton:compile(r, g, b))
	shader:send("h", love.graphics.getHeight())

	print(scale, size, size / scale)
end

function love.load()
	slider = {value = 0}
	scale = 1
	image = love.graphics.newCanvas()
	image:setFilter("nearest", "nearest")

	automaton = Glitch()

	acc = 0
	t = 0
	local w, h = love.graphics.getWidth(), love.graphics.getHeight()

	collider = Collider()
	nodes = {}

	for i=1,10 do
		local index = ((i - 1) % #BINARY) + 1
		nodes[i] = Node(BINARY[index], math.random() * 512, math.random() * 512, collider)
	end

	local walls = {}
	
	local wall
	wall = Rectangle(512, 100)
	wall:moveTo(256, -50)
	table.insert(walls, wall)

	wall = Rectangle(512, 100)
	wall:moveTo(256, 512+50)
	table.insert(walls, wall)

	wall = Rectangle(100, 512)
	wall:moveTo(-50, 256)
	table.insert(walls, wall)

	wall = Rectangle(100, 512)
	wall:moveTo(512+50, 256)
	table.insert(walls, wall)

	for i, wall in ipairs(walls) do
		wall.self = wall
		wall.static = true
		collider:addShape(wall)
		collider:setPassive(wall)

		table.insert(nodes, wall)
	end

	r = Terminal({name="r"}, 128, 512-12.5, collider)
	g = Terminal({name="g"}, 256, 512-12.5, collider)
	b = Terminal({name="b"}, 388, 512-12.5, collider)

	table.insert(nodes, Input(VARIABLES[2], 128, 12.5, collider))
	table.insert(nodes, Input(VARIABLES[3], 256, 12.5, collider))
	table.insert(nodes, Input(VARIABLES[1], 388, 12.5, collider))

	table.insert(nodes, r)
	table.insert(nodes, g)
	table.insert(nodes, b)

	refresh()

	collider:setCallbacks(function (dt, shape_a, shape_b, dx, dy)
		if not shape_a.self.static and not shape_b.self.static then
			shape_a:move( dx/2,  dy/2)
			shape_b:move(-dx/2, -dy/2)
		elseif not shape_a.self.static then
			shape_a:move(dx, dy)
		elseif not shape_b.self.static then
			shape_b:move(-dx, -dy)
		end
	end)
end

function love.update(dt)
	acc = acc + dt

	if move then
		local x, y = love.mouse.getPosition()

		x = math.max(0, math.min(x, 512))
		y = math.max(0, math.min(y, 512))

		move:move(x, y)
	end

	gui.Slider{info = slider}
	scale = math.pow(2, math.floor(slider.value*8))

	while acc >= 0.01 do
		acc = acc - 0.01
		dt = 0.01

		t = (t + dt) % 1

		collider:update(dt)
		shader:send("t", t)
		shader:send("s", size / scale)
	end
end

function love.draw()
	--[[
	local f = 8

	love.graphics.setPixelEffect()
	for y=0,63 do
		for x=0,63 do
			love.graphics.setColor(automaton:compute(x / 63, y / 63, t))
			love.graphics.rectangle("fill", 512 + x * f, y * f, f, f)
		end
	end
	]]

	image:clear()
	love.graphics.setColor(255, 255, 255, 255)
	love.graphics.setPixelEffect(shader)
	image:renderTo(function () 
		love.graphics.rectangle("fill", 0, 0, size / scale, size / scale)
	end)

	love.graphics.setPixelEffect()
	love.graphics.draw(image, 0, 0, 0, scale, scale)

	for i, node in pairs(nodes) do
		if drag and drag.ancestors[node] then
			love.graphics.setColor(255, 128, 128, 255)
		else
			love.graphics.setColor(255, 255, 255, 255)
		end

		node:draw()
	end

	if drag then
		local x, y = love.mouse.getPosition()
		love.graphics.line(x, y, drag.shape:center())
	end

	gui.core.draw()
end

function love.mousepressed(x, y, button)
	for i, shape in pairs(collider:shapesAt(x, y)) do
		if shape.type == "node" then
			if button == "l" and not shape.self.static then
				move = shape.self
			elseif button == "r" and shape.self.outputs then
				drag = shape.self
			end
		end
	end
end

function love.mousereleased(x, y, button)
	if button == "l" then move = nil end
	if button == "r" then
		for i, shape in pairs(collider:shapesAt(x, y)) do
			if drag and shape.type == "node" and not drag.ancestors[shape.self] and shape.self.inputs then
				drag:into(shape.self)
				refresh()
				break
			end
		end

		drag = nil
	end
end
